// -----------------------------------------------------------------------------
// Copyright 2011-2012 Patrick Näf (herzbube@herzbube.ch)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// -----------------------------------------------------------------------------


// Forward declarations
@class GoPlayer;
@class GoPoint;


// -----------------------------------------------------------------------------
/// @brief The GoMove class represents a move made by one of the players.
///
/// @ingroup go
///
/// A GoMove object always has a type(); the different types of moves are
/// enumerated by #GoMoveType. A GoMove object is always associated with the
/// player who made the move. The player object can be queried for the color of
/// the move.
///
/// If a GoMove object is of type #GoMoveTypePlay it also has an associated
/// GoPoint object which registers where the stone was placed. The GoPoint
/// object is assigned (soon) after construction.
///
/// @note Assignment of a GoPoint object is a non-trivial operation that
/// triggers the mechanism for placing a stone.
///
/// GoMove objects are interlinked with their predecessor (previous()) and
/// successor (next()) GoMove object. This represents the fact that a game
/// can be seen as a series of moves.
///
/// GoMove has undo support. Invoking undo() reverts the board to the state it
/// had before the GoMove was played. Invoking undo() also removes references
/// from/to the predecessor GoMove, which usually causes the GoMove to be
/// deallocated.
// -----------------------------------------------------------------------------
@interface GoMove : NSObject <NSCoding>
{
}

+ (GoMove*) move:(enum GoMoveType)type by:(GoPlayer*)player after:(GoMove*)move;
- (void) undo;

/// @brief The type of this GoMove object.
@property(nonatomic, assign, readonly) enum GoMoveType type;
/// @brief The player who made this GoMove.
@property(nonatomic, retain, readonly) GoPlayer* player;
/// @brief The GoPoint object registering where the stone was placed for this
/// GoMove. Is nil if this GoMove is @e not a #GoMoveTypePlay.
@property(nonatomic, assign) GoPoint* point;
/// @brief The predecessor to this GoMove object. nil if this is the first move
/// of the game.
@property(nonatomic, assign, readonly) GoMove* previous;  // do not retain, otherwise there would be a retain cycle
/// @brief The successor to this GoMove object. nil if this is the last move
/// of the game.
@property(nonatomic, retain, readonly) GoMove* next;      // retain here, making us the parent, and next the child
/// @brief Keeps track of stones that were captured by this move.
///
/// If not empty, the array contains an unordered list of GoPoint objects. Also,
/// if several stone groups were captured, the GoPoint objects do not form a
/// single contiguous region.
@property(nonatomic, retain, readonly) NSArray* capturedStones;
/// @brief Is true if this move was generated by the computer during interactive
/// play.
///
/// This flag can become true either because it was a computer player's turn or
/// because a human player requested that the computer make his or her move.
///
/// This flag remains false if a human player made his or her move, or if this
/// move was generated non-interactively (e.g. when a game was loaded from an
/// archive).
@property(nonatomic, assign) bool computerGenerated;

@end
